using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace InputField {

	public class InputFieldGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		public InputFieldGame() {
			graphics = new GraphicsDeviceManager(this);
			graphics.PreferMultiSampling = true;
			graphics.PreferredBackBufferWidth = 1280;
			graphics.PreferredBackBufferHeight = 720;
			Content.RootDirectory = "Content";
			Components.Add(new TextFieldComponent(this));
		}
		protected override void Update(GameTime gameTime) {
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();
			base.Update(gameTime);
		}
		protected override void Draw(GameTime gameTime) {
			GraphicsDevice.Clear(Color.CornflowerBlue);
			base.Draw(gameTime);
		}
	}

	public class TextFieldComponent : DrawableGameComponent {
		SpriteBatch spriteBatch;
		SpriteFont font;
		Keys[] previousPressedKeys;
		Keys[] currentPressedKeys = new Keys[0];
		public int CaretBlinkSpan = 30;
		int caretBlinkTime;
		readonly StringBuilder text = new StringBuilder();
		bool caretVisible;
		readonly InputField field = new InputField();
		readonly Dictionary<Keys, KeyBind> dictionary = new Dictionary<Keys, KeyBind>();
		public int KeyRepeatFirstSpan = 20;
		public int KeyRepeatSpan = 5;
		int keyRepeatTime;
		Keys lastKey;

		public TextFieldComponent(Game game) : base(game) { }

		public override void Initialize() {
			keyRepeatTime = KeyRepeatFirstSpan;
			foreach (var bind in binds) {
				dictionary.Add(bind.Key, bind);
			}
			base.Initialize();
		}
		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			font = Game.Content.Load<SpriteFont>("DefaultFont");
		}

		interface Control {
			StringBuilder Text { get; }
			int CurrentIndex { get; set; }
			int MaximumLength { get; set; }
			bool IsShiftPressed { get; set; }
			void ClampText();
		}
		class InputField : Control {
			public InputField() { Text = new StringBuilder(); MaximumLength = 16; }
			public StringBuilder Text { get; protected set; }
			public int CurrentIndex { get; set; }
			public int MaximumLength { get; set; }
			public bool IsShiftPressed { get; set; }
			public void ClampText() {
				if (Text.Length > MaximumLength) {
					Text.Remove(MaximumLength, Text.Length - MaximumLength);
					CurrentIndex = Math.Min(MaximumLength, CurrentIndex);
				}
			}
		}

		interface KeyBind {
			Keys Key { get; }
			void Apply(Control control);
		};
		class Alphabet : KeyBind {
			public readonly string Text;
			public readonly string ShiftText;
			public Alphabet(Keys key) {
				Key = key; Text = key.ToString();
			}
			public Alphabet(Keys key, char defaultValue, char shiftValue) {
				Key = key; Text = new string(defaultValue, 1);
				Key = key; ShiftText = new string(shiftValue, 1);
			}
			public Alphabet(Keys key, string text) {
				Key = key; Text = text;
			}
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				var text = control.IsShiftPressed ? ShiftText : Text;
				if (text != null) {
					control.Text.Insert(control.CurrentIndex, text);
					control.CurrentIndex += text.Length;
					control.ClampText();
				}
			}
		}
		class Enter : KeyBind {
			public Enter(Keys key) { Key = key; }
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				/*
				// sT|[gꍇiʓrALbg̏̕ύXKvj
				control.Text.Insert(control.CurrentIndex, '\n');
				control.CurrentIndex++;
				control.ClampText();
				*/
			}
		}
		class Back : KeyBind {
			public Back(Keys key) { Key = key; }
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				if (control.CurrentIndex >= 1 && control.CurrentIndex <= control.Text.Length) {
					control.Text.Remove(--control.CurrentIndex, 1);
				}
			}
		}
		class Delete : KeyBind {
			public Delete(Keys key) { Key = key; }
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				if (control.CurrentIndex >= 0 && control.CurrentIndex < control.Text.Length) {
					control.Text.Remove(control.CurrentIndex, 1);
				}
			}
		}
		class Left : KeyBind {
			public Left(Keys key) { Key = key; }
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				if (control.CurrentIndex > 0) {
					control.CurrentIndex--;
				}
			}
		}
		class Right : KeyBind {
			public Right(Keys key) { Key = key; }
			public Keys Key { get; protected set; }
			public void Apply(Control control) {
				if (control.CurrentIndex < control.Text.Length) {
					control.CurrentIndex++;
				}
			}
		}

		static readonly KeyBind[] binds = new KeyBind[]
		{
			new Enter(Keys.Enter),
			new Back(Keys.Back),
			new Delete(Keys.Delete),
			new Left(Keys.Left),
			new Right(Keys.Right),
			new Alphabet(Keys.D1, '1', '!'),
			new Alphabet(Keys.D2, '2', '"'),
			new Alphabet(Keys.D3, '3', '#'),
			new Alphabet(Keys.D4, '4', '$'),
			new Alphabet(Keys.D5, '5', '%'),
			new Alphabet(Keys.D6, '6', '&'),
			new Alphabet(Keys.D7, '7', '\''),
			new Alphabet(Keys.D8, '8', '('),
			new Alphabet(Keys.D9, '9', ')'),
			new Alphabet(Keys.D0, "0"),
			new Alphabet(Keys.Space, ' ', ' '),
			new Alphabet(Keys.A, 'a', 'A'),
			new Alphabet(Keys.B, 'b', 'B'),
			new Alphabet(Keys.C, 'c', 'C'),
			new Alphabet(Keys.D, 'd', 'D'),
			new Alphabet(Keys.E, 'e', 'E'),
			new Alphabet(Keys.F, 'f', 'F'),
			new Alphabet(Keys.G, 'g', 'G'),
			new Alphabet(Keys.H, 'h', 'H'),
			new Alphabet(Keys.I, 'i', 'I'),
			new Alphabet(Keys.J, 'j', 'J'),
			new Alphabet(Keys.K, 'k', 'K'),
			new Alphabet(Keys.L, 'l', 'L'),
			new Alphabet(Keys.M, 'm', 'M'),
			new Alphabet(Keys.N, 'n', 'N'),
			new Alphabet(Keys.O, 'o', 'O'),
			new Alphabet(Keys.P, 'p', 'P'),
			new Alphabet(Keys.Q, 'q', 'Q'),
			new Alphabet(Keys.R, 'r', 'R'),
			new Alphabet(Keys.S, 's', 'S'),
			new Alphabet(Keys.T, 't', 'T'),
			new Alphabet(Keys.U, 'u', 'U'),
			new Alphabet(Keys.V, 'v', 'V'),
			new Alphabet(Keys.W, 'w', 'W'),
			new Alphabet(Keys.X, 'x', 'X'),
			new Alphabet(Keys.Y, 'y', 'Y'),
			new Alphabet(Keys.Z, 'z', 'Z'),
			new Alphabet(Keys.OemMinus, '-', '='),
			new Alphabet(Keys.OemQuotes, '^', '~'),
			new Alphabet(Keys.OemTilde, '@','`'),
			new Alphabet(Keys.OemBackslash, '\\', '_'),
			new Alphabet(Keys.OemPipe, '\\', '|'),
			new Alphabet(Keys.OemPlus, ';', '+'),
			new Alphabet(Keys.OemOpenBrackets, '[', '{'),
			new Alphabet(Keys.OemCloseBrackets, ']', '}'),
			new Alphabet(Keys.OemSemicolon, ':', '*'),
			new Alphabet(Keys.OemComma, ',', '<'),
			new Alphabet(Keys.OemPeriod, '.', '>'),
			new Alphabet(Keys.OemQuestion, '/', '?'),
		};

		public override void Update(GameTime gameTime) {
			KeyboardState state = Keyboard.GetState();
			previousPressedKeys = currentPressedKeys;
			currentPressedKeys = state.GetPressedKeys();

			field.IsShiftPressed = state.IsKeyDown(Keys.LeftShift) || state.IsKeyDown(Keys.RightShift);

			foreach (Keys key in currentPressedKeys) {
				bool hit;
				if (!previousPressedKeys.Contains(key)) {
					keyRepeatTime = KeyRepeatFirstSpan;
					hit = true;
					lastKey = key;
				} else {
					if (lastKey == key && --keyRepeatTime <= 0) {
						hit = true;
						keyRepeatTime = KeyRepeatSpan;
					} else {
						hit = false;
					}
				}
				if (hit) {
					KeyBind bind;
					if (dictionary.TryGetValue(key, out bind)) {
						bind.Apply(field);
					}
				}
			}

			if (++caretBlinkTime >= CaretBlinkSpan) {
				caretBlinkTime = 0;
				caretVisible = !caretVisible;
			}

			base.Update(gameTime);
		}

		Vector2 location = new Vector2(0x80, 0x100);

		public override void Draw(GameTime gameTime) {
			var caretLocation = font.MeasureString(field.Text.ToString().Substring(0, field.CurrentIndex));
			caretLocation.Y = 0;
			Vector2.Add(ref caretLocation, ref location, out caretLocation);

			spriteBatch.Begin();
			Vector2 shadowLocation = location;
			shadowLocation.X++;
			shadowLocation.Y++;
			spriteBatch.DrawString(font, field.Text, shadowLocation, Color.Black);
			spriteBatch.DrawString(font, field.Text, location, Color.White);
			if (caretVisible) spriteBatch.DrawString(font, "[", caretLocation, Color.Red);
			spriteBatch.End();

			base.Draw(gameTime);
		}
	}
}
